from allennlp_models.structured_prediction.predictors.biaffine_dependency_parser import (
    BiaffineDependencyParserPredictor,
)
from allennlp_models.structured_prediction.predictors.constituency_parser import (
    ConstituencyParserPredictor,
)
from allennlp_models.structured_prediction.predictors.openie import OpenIePredictor
from allennlp_models.structured_prediction.predictors.srl import SemanticRoleLabelerPredictor
