from allennlp_models.structured_prediction.dataset_readers.penn_tree_bank import (
    PennTreeBankConstituencySpanDatasetReader,
)
from allennlp_models.structured_prediction.dataset_readers.semantic_dependencies import (
    SemanticDependenciesDatasetReader,
)
from allennlp_models.structured_prediction.dataset_readers.srl import SrlReader
from allennlp_models.structured_prediction.dataset_readers.universal_dependencies import (
    UniversalDependenciesDatasetReader,
)
