from allennlp_models.rc.models.bidaf_ensemble import BidafEnsemble
from allennlp_models.rc.models.bidaf import BidirectionalAttentionFlow
from allennlp_models.rc.models.dialog_qa import DialogQA
from allennlp_models.rc.models.naqanet import NumericallyAugmentedQaNet
from allennlp_models.rc.models.qanet import QaNet
from allennlp_models.rc.models.transformer_qa import TransformerQA
from allennlp_models.rc.models.utils import get_best_span
